
blp_date_string(date::Date) :: String = Dates.format(date, dateformat"yyyymmdd")
blpstring(str::AbstractString) = str
blpstring(date::Date) = blp_date_string(date)

function SchemaElementDefinition(element::Element)
    return SchemaElementDefinition(blpapi_Element_definition(element.handle))
end

function is_array(::Type{Element{A,D}}) :: Bool where {A,D}
    A
end

function is_array(::Element{A,D}) :: Bool where {A,D}
    A
end

@generated function get_child_element(element::AbstractElement{A,D}, index) where {A,D}
    if is_complex_datatype(D)

        @assert !A "get_child_element not allowed for array element."

        if index <: Integer
            # index into element using an Integer as key
            return quote
                result_element_handle_ref = Ref{Ptr{Cvoid}}(C_NULL)
                err = blpapi_Element_getElementAt(element.handle, result_element_handle_ref, index - 1)
                error_check(err, "Failed to get element at index $index from $(element.name)")
                return Element(result_element_handle_ref[], element)
            end

        elseif index <: AbstractString
            # index into element using a String as key
            return quote
                result_element_handle_ref = Ref{Ptr{Cvoid}}(C_NULL)
                err = blpapi_Element_getElement(element.handle, result_element_handle_ref, pointer(index), C_NULL)
                error_check(err, "Failed to get element $index from $(element.name)")
                return Element(result_element_handle_ref[], element)
            end
        else
            error("index data type $index not supported")
        end
    end

    error("get_child_element not allowed for data type $D.")
end

@generated function Base.getindex(element::AbstractElement{false,D}, index::AbstractString) where {D}

    # index <: AbstractString
    # used to index into child elements of a complex element that is not an array
    if is_complex_datatype(D)
        return quote
            return get_child_element(element, index)
        end
    end

    error("Base.getindex not implemented for element with datatype $D with is_array = $A.")
end

has_name(element::Element, name) = has_name(element.name, name)
get_name(element::Element) = element.name

function get_choice(element::AbstractElement{false,BLPAPI_DATATYPE_CHOICE})
    result_element_handle_ref = Ref{Ptr{Cvoid}}(C_NULL)
    err = blpapi_Element_getChoice(element.handle, result_element_handle_ref)
    error_check(err, "Failed to get choice from $(get_name(element))")
    return Element(result_element_handle_ref[], element)
end

Base.getindex(element::AbstractElement{false,BLPAPI_DATATYPE_CHOICE}) = get_choice(element)

@generated function has_child_element(element::AbstractElement{false,D}, name::AbstractString) :: Bool where {D}
    if is_complex_datatype(D)
        return quote
            return blpapi_Element_hasElement(element.handle, pointer(name), C_NULL) != 0
        end
    end

    error("has_child_element not implemented for element with datatype $D.")
end

function Base.haskey(element::AbstractElement{false,D}, name::AbstractString) :: Bool where {D}
    has_child_element(element, name)
end

struct ChildElementsIterator{T<:AbstractElement{false,BLPAPI_DATATYPE_SEQUENCE}}
    element::T
    num_elements::Csize_t
end

function each_child_element(element::T) where {T<:AbstractElement{false, BLPAPI_DATATYPE_SEQUENCE}}
    ChildElementsIterator(element, blpapi_Element_numElements(element.handle))
end

Base.isempty(itr::ChildElementsIterator) = itr.num_elements == 0
Base.length(itr::ChildElementsIterator) = itr.num_elements

function Base.iterate(itr::ChildElementsIterator)
    if isempty(itr)
        return nothing
    else
        # (current_element, next_index)
        return (get_child_element(itr.element, UInt64(1)), UInt64(2))
    end
end

function Base.iterate(itr::ChildElementsIterator, state::UInt64)
    if state > itr.num_elements
        return nothing
    else
        return (get_child_element(itr.element, state), state + UInt64(1))
    end
end

"""
    get_element_value(element, [index])

Returns the Julia value for `element`.

`element` cannot be a `BLPAPI_DATATYPE_SEQUENCE` or `BLPAPI_DATATYPE_CHOICE`.
"""
@generated function get_element_value(element::AbstractElement{A,D}) where {A,D}
    if A
        # retrieves a vector
        return quote
            num_values = blpapi_Element_numValues(element.handle)

            result = Vector()
            for i in 1:num_values
                push!(result, get_element_value(element, i))
            end

            return result
        end
    else
        if is_complex_datatype(D)
            error("get_element_value: Can't get element value from a complex element that is not an array.")
        end

        # retrieves a scalar
        return quote
            return get_element_value(element, 1)
        end
    end
end

@generated function get_element_value(element::AbstractElement{A,D}, index::Integer) where {A,D}
    if D == BLPAPI_DATATYPE_BOOL
        parse_value_block = quote
            buffer_ref = Ref{Cint}()
            err = blpapi_Element_getValueAsBool(element.handle, buffer_ref, index-1)
            error_check(err, "Failed to get Bool element value")
            return buffer_ref[] != 0
        end

    elseif D == BLPAPI_DATATYPE_CHAR
        parse_value_block = quote
            buffer_ref = Ref{Cchar}()
            err = blpapi_Element_getValueAsChar(element.handle, buffer_ref, index-1)
            error_check(err, "Failed to get Char element value")
            return Char(buffer_ref[])
        end

    elseif D == BLPAPI_DATATYPE_INT32
        parse_value_block = quote
            buffer_ref = Ref{Int32}()
            err = blpapi_Element_getValueAsInt32(element.handle, buffer_ref, index-1)
            error_check(err, "Failed to get Int32 element value")
            return buffer_ref[]
        end

    elseif D == BLPAPI_DATATYPE_INT64
        parse_value_block = quote
            buffer_ref = Ref{Int64}()
            err = blpapi_Element_getValueAsInt64(element.handle, buffer_ref, index-1)
            error_check(err, "Failed to get Int32 element value")
            return buffer_ref[]
        end

    elseif D == BLPAPI_DATATYPE_FLOAT32
        parse_value_block = quote
            buffer_ref = Ref{Float32}()
            err = blpapi_Element_getValueAsFloat32(element.handle, buffer_ref, index-1)
            error_check(err, "Failed to get Float32 element value")
            return buffer_ref[]
        end

    elseif D == BLPAPI_DATATYPE_FLOAT64
        parse_value_block = quote
            buffer_ref = Ref{Float64}()
            err = blpapi_Element_getValueAsFloat64(element.handle, buffer_ref, index-1)
            error_check(err, "Failed to get Float64 element value")
            return buffer_ref[]
        end

    elseif D == BLPAPI_DATATYPE_STRING
        parse_value_block = quote
            buffer_ref = Ref{Cstring}()
            err = blpapi_Element_getValueAsString(element.handle, buffer_ref, index-1)
            error_check(err, "Failed to get String element value")
            return unsafe_string(buffer_ref[])
        end

    elseif D == BLPAPI_DATATYPE_DATE
        parse_value_block = quote
            buffer_ref = Ref{BLPDateTime}()
            err = blpapi_Element_getValueAsDatetime(element.handle, buffer_ref, index-1)
            error_check(err, "Failed to get Date element value")
            return Dates.Date(buffer_ref[])
        end

    elseif D == BLPAPI_DATATYPE_TIME
        parse_value_block = quote
            buffer_ref = Ref{BLPDateTime}()
            err = blpapi_Element_getValueAsDatetime(element.handle, buffer_ref, index-1)
            error_check(err, "Failed to get Date element value")
            return Dates.Time(buffer_ref[])
        end

    elseif D == BLPAPI_DATATYPE_DATETIME
        parse_value_block = quote
            buffer_ref = Ref{BLPDateTime}()
            err = blpapi_Element_getValueAsDatetime(element.handle, buffer_ref, index-1)
            error_check(err, "Failed to get Date element value")
            #return DateTime(buffer_ref[]) # TODO: use TimeZones.
            return buffer_ref[]
        end

    elseif D == BLPAPI_DATATYPE_SEQUENCE
        # get_element_value can be used on sequence only if it is an array
        @assert A "get_element_value: Can't get element value from datatype $D that is not an array (is_array = $A)."

        parse_value_block = quote
            buffer_ref = Ref{Ptr{Cvoid}}(C_NULL)
            err = blpapi_Element_getValueAsElement(element.handle, buffer_ref, index-1)
            error_check(err, "Failed to get Date element value")
            return Element(buffer_ref[], element)
        end

    elseif D == BLPAPI_DATATYPE_ENUMERATION
        parse_value_block = quote
            buffer_ref = Ref{Ptr{Cvoid}}(C_NULL)
            err = blpapi_Element_getValueAsName(element.handle, buffer_ref, index-1)
            error_check(err, "Failed to get BLPName element value from Enumeration")
            return Symbol(BLPName(buffer_ref[]))
        end

    else
        error("get_element_value: support for datatype $D not implemented.")
    end

    return quote
        if is_null_element(element)
            return missing
        else
            $parse_value_block
        end
    end
end

@generated function set_element_value!(element::AbstractElement{A,D}, val::V, index::Integer) where {A,D,V}
    if D == BLPAPI_DATATYPE_STRING
        return quote
            err = blpapi_Element_setValueString(element.handle, blpstring(val), index)
            error_check(err, "Failed to push value $val to element $(get_name(element))")
        end

    elseif D == BLPAPI_DATATYPE_BOOL
        @assert val == Bool "Can't set value of type $val to element of type $D."
        return quote
            err = blpapi_Element_setValueBool(element.handle, val, index)
            error_check(err, "Failed to push value $val to element $(get_name(element))")
        end

    elseif D == BLPAPI_DATATYPE_INT32
        @assert val <: Integer
        return quote
            err = blpapi_Element_setValueInt32(element.handle, Int32(val), index)
            error_check(err, "Failed to push value $val to element $(get_name(element))")
        end

    elseif D == BLPAPI_DATATYPE_INT64
        @assert val <: Integer
        return quote
            err = blpapi_Element_setValueInt64(element.handle, Int64(val), index)
            error_check(err, "Failed to push value $val to element $(get_name(element))")
        end

    elseif D == BLPAPI_DATATYPE_FLOAT32
        @assert val <: Number
        return quote
            err = blpapi_Element_setValueFloat32(element.handle, Float32(val), index)
            error_check(err, "Failed to push value $val to element $(get_name(element))")
        end

    elseif D == BLPAPI_DATATYPE_FLOAT64
        @assert val <: Number
        return quote
            err = blpapi_Element_setValueFloat64(element.handle, Float64(val), index)
            error_check(err, "Failed to push value $val to element $(get_name(element))")
        end

    elseif D == BLPAPI_DATATYPE_CHAR
        @assert val == Char
        return quote
            err = blpapi_Element_setValueChar(element.handle, val, index)
            error_check(err, "Failed to push value $val to element $(get_name(element))")
        end

    elseif D == BLPAPI_DATATYPE_ENUMERATION
        # for enums, we need to check the value type informed by the element's schema
        # and also if the value is allowed for the field
        return quote
            schema = SchemaElementDefinition(element)
            schema_type = schema.schema_type
            @assert isa(schema_type, EnumerationSchemaTypeDefinition)

            if schema_type.enumeration.datatype == BLPAPI_DATATYPE_STRING
                val_str = blpstring(val)
                @assert is_value_allowed(element, val_str) "Unvalid value for enum `$(get_name(element))`: `$val_str`. Options are: $(list_enum_options(element))"

                err = blpapi_Element_setValueString(element.handle, val_str, index)
                error_check(err, "Failed to push value $val_str to element $(get_name(element))")
            else
                error("enumeration datatype not supported: $(schema_type.enumeration.datatype)")
            end
        end

    elseif D == BLPAPI_DATATYPE_DATETIME || D == BLPAPI_DATATYPE_DATE
        @assert val == BLPDateTime || val == DateTime || val == Date

        return quote
            err = blpapi_Element_setValueDatetime(element.handle, BLPDateTime(val), index)
            error_check(err, "Failed to push value $val to element $(get_name(element))")
        end
    end

    error("set_element_value! not implemented for datatype $D.")
end

# allows `element[] = val`, when element is not an array
function Base.setindex!(element::Element{false,D}, val::V) where {D,V}
    set_element_value!(element, val, 0)
end

function is_value_allowed(element::Element{A, BLPAPI_DATATYPE_ENUMERATION}, val) where {A}
    schema = SchemaElementDefinition(element)
    schema_type = schema.schema_type
    @assert isa(schema_type, EnumerationSchemaTypeDefinition)

    # check wether val is allowed for this enum
    val_is_allowed = false
    for enum_const in schema_type.enumeration.list
        if enum_const.value == val
            val_is_allowed = true
            break
        end
    end

    return val_is_allowed
end

function list_enum_options(element::Element{A, BLPAPI_DATATYPE_ENUMERATION}) where {A}
    schema = SchemaElementDefinition(element)
    schema_type = schema.schema_type
    @assert isa(schema_type, EnumerationSchemaTypeDefinition)
    return [ enum_const.value for enum_const in schema_type.enumeration.list ]
end

function Base.push!(element::Element{true,D}, val::V) where {D,V}
    set_element_value!(element, val, BLPAPI_ELEMENT_INDEX_END)
end

function Base.push!(element::Element, vals...)
    for val in vals
        push!(element, val)
    end
end

Base.append!(element::Element{true}, vals::Vector) = push!(element, vals...)

# allows `element["child_element_name"] = val`, when element is complex and not an array
@generated function Base.setindex!(element::Element{false,D}, val::V, name::AbstractString) where {D,V}
    if is_complex_datatype(D)
        return quote
            child_element = element[name]
            child_element[] = val
            nothing
        end
    end

    error("setindex! not implemented for datatype $D.")
end

function is_null_value(element::AbstractElement, index::Integer)
    rc = blpapi_Element_isNullValue(element.handle, index-1)

    if rc != 0 && rc != 1
        error("Unexpected result of blpapi_Element_isNullValue on element of type $(typeof(element)): $rc")
    end

    return rc != 0
end

function is_null_element(element::AbstractElement)
    rc = blpapi_Element_isNull(element.handle)
    return rc != 0
end

function as_dict!(dict::Dict, root::Element{A,D}) where {A,D}

    if is_choice_datatype(D)
        as_dict!(dict, get_choice(root))

    elseif D == BLPAPI_DATATYPE_SEQUENCE

        # vector of sequences
        if A
            elements_vec = Vector()
            root_name_symbol = Symbol(get_name(root))
            dict[root_name_symbol] = elements_vec
            for el in get_element_value(root)
                el_dict = Dict()
                as_dict!(el_dict, el)
                if length(el_dict) == 1 && haskey(el_dict, root_name_symbol)
                    push!(elements_vec, el_dict[root_name_symbol])
                else
                    push!(elements_vec, el_dict)
                end
            end
        else
            # a single sequence
            child_dict = Dict()
            dict[Symbol(get_name(root))] = child_dict
            for child_element in each_child_element(root)
                as_dict!(child_dict, child_element)
            end
        end

    else
        @assert is_simple_datatype(D)
        dict[Symbol(get_name(root))] = get_element_value(root)
    end
end

function Base.Dict(element::AbstractElement)
    dict = Dict()
    as_dict!(dict, element)
    return dict
end

function push_named_tuples!(result::T, element_vec::Element{true, BLPAPI_DATATYPE_SEQUENCE}) where {T<:Vector}
    tuple_keys = nothing

    for element in get_element_value(element_vec)

        if tuple_keys == nothing
            tuple_keys = Tuple([ Symbol(get_name(child_element)) for child_element in each_child_element(element) ])
        else
            @assert tuple_keys == Tuple([ Symbol(get_name(child_element)) for child_element in each_child_element(element) ])
        end

        tuple_values = Tuple([ get_element_value(child_element) for child_element in each_child_element(element) ])
        push!(result, (; zip(tuple_keys, tuple_values)...)) # trick based on the docstring for NamedTuple
    end
end

to_named_tuple(element::Element{false, BLPAPI_DATATYPE_SEQUENCE}) = (; [ (Symbol(get_name(child_element)), get_element_value(child_element)) for child_element in each_child_element(element) ]...)
